// BinaryMeshFile.cpp: implementation of the BinaryMeshFile class.
//
//////////////////////////////////////////////////////////////////////

#include "BinaryMeshFile.h"
#include "TextureFactory.h"
#include "FileNotFoundException.h"
#include "glut.h"
#include "glerr.h"
#include "Math.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


BinaryMeshFile::BinaryMeshFile(char filename[])
{
  int i;
  char buffer[80];

  minx=miny=minz=maxx=maxy=maxz=0.0f;
  in.open(filename, ios::in| ios::binary );
  if(!in)
    throw FileNotFoundException(filename);
  in.read((char *) &radius, sizeof(radius) );
  readCoordinate(center);
  readCoordinate(extents);
  in.read((char *) &numtextures, sizeof(numtextures) );
  textures = new Texture *[numtextures];
  for(i=0; i < numtextures; i++)
  {
    int stringlength;
    in.read((char *) &stringlength, sizeof(stringlength) );
    in.read(buffer, stringlength);
    buffer[stringlength] = '\0';
    textures[i] = TextureFactory::getTexture(buffer);
  }
  in.read((char *) &numfaces, sizeof(numfaces) );
  faces = new Face [numfaces + numtextures];
  for(i=0; i < numfaces + numtextures - 1; i++)
  {
     readFace(i);
  }
  mycollision_detection = newCollisionModel3D();
  initCollisionDetection(mycollision_detection, Coordinate(0.0f, 0.0f, 0.0f) );
  mycollision_detection->finalize();
//  initExtents();
}

BinaryMeshFile::BinaryMeshFile(char filename [ ], Color transparent)
{
  int i;
  char buffer[80];

  minx=miny=minz=maxx=maxy=maxz=0.0f;
  in.open(filename, ios::in| ios::binary );
  if(!in)
    throw FileNotFoundException(filename);
  in.read((char *) &radius, sizeof(radius) );
  readCoordinate(center);
  readCoordinate(extents);
  in.read((char *) &numtextures, sizeof(numtextures) );
  textures = new Texture *[numtextures];
  for(i=0; i < numtextures; i++)
  {
    int stringlength;
    in.read((char *) &stringlength, sizeof(stringlength) );
    in.read(buffer, stringlength);
    buffer[stringlength] = '\0';
    textures[i] = TextureFactory::getTransparentTexture(buffer, transparent);
  }
  in.read((char *) &numfaces, sizeof(numfaces) );
  faces = new Face [numfaces + numtextures];
  for(i=0; i < numfaces + numtextures - 1; i++)
  {
     readFace(i);
  }
  mycollision_detection = newCollisionModel3D();
  initCollisionDetection(mycollision_detection, Coordinate(0.0f, 0.0f, 0.0f) );
  mycollision_detection->finalize();
//  initExtents();
}


BinaryMeshFile::~BinaryMeshFile()
{
  delete [] faces;
  delete [] textures;
}

Coordinate BinaryMeshFile::getExtents()
{
  return extents;
}

Coordinate BinaryMeshFile::getcenter()
{
  return center;
}

float BinaryMeshFile::getradius()
{
  return radius;
}


void BinaryMeshFile::render()
{
  int currentpoint;
  int currentface;
  bool lastwastriangle=true;
  int lasttexture = 0;
  int i;

  glEnable(GL_TEXTURE_2D);
  glDisable(GL_COLOR_MATERIAL);
  glColor3f(1.0f, 1.0f, 1.0f);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  textures[0]->render();
//  glColor3f(1.0f, 0.0f, 0.0f);
  glBegin(GL_TRIANGLES);
  currentpoint = 0;

  for(currentface =0; currentface < numfaces+ numtextures - 1; currentface++)
  {
     Face & thisface = faces[currentface];

     if(thisface.numpoints == 3)
     {
/*        if(whichtexture[currentface] != lasttexture)
        {
           glEnd();
           textures[ whichtexture[currentface]]->render();
           glBegin(GL_TRIANGLES);
        }
        else*/ if(lastwastriangle == false)
        {
           glEnd();
           glBegin(GL_TRIANGLES);
        }

        for(i=0; i < 3; i++)
        {
           glTexCoord2f(thisface.texels[i].s, thisface.texels[i].t);
  	        glNormal3f(thisface.normals[i].x, thisface.normals[i].y, thisface.normals[i].z);
           glVertex3f(thisface.points[i].x , thisface.points[i].y , thisface.points[i].z);
        }
        lastwastriangle = true;
     }
     else if(thisface.numpoints == 4)
     {
/*        if(whichtexture[currentface] != lasttexture)
        {
           glEnd();
           textures[ whichtexture[currentface]]->render();
           glBegin(GL_QUADS);
        }
        else*/ if(lastwastriangle == true)
        {
           glEnd();
           glBegin(GL_QUADS);
        }
        for(i=0; i < 4; i++)
        {
           glTexCoord2f(thisface.texels[i].s, thisface.texels[i].t);
  	        glNormal3f(thisface.normals[i].x, thisface.normals[i].y, thisface.normals[i].z);
           glVertex3f(thisface.points[i].x , thisface.points[i].y , thisface.points[i].z);
        }

        lastwastriangle = false;
     }
     else
     {
       // texture swap
       glEnd();
       int texturenum = (int) thisface.points[0].x;

       textures[texturenum]->render();
       if(lastwastriangle)
       {
           glBegin(GL_TRIANGLES);
       }
       else
       {
           glBegin(GL_QUADS);
       }
     }
  }
  glEnd();
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_COLOR_MATERIAL);
  GLERRORCHECK
}

void BinaryMeshFile::readCoordinate(Coordinate & coord)
{
  in.read((char *) &coord.x, sizeof(coord.x) );
  in.read((char *) &coord.y, sizeof(coord.y) );
  in.read((char *) &coord.z, sizeof(coord.z) );
}

void BinaryMeshFile::readFace(int currentface)
{
   Face & thisface = faces[currentface];
   int i;
   int numpoints;

   in.read((char *) &numpoints, sizeof(numpoints) );

   if(numpoints == 0) // texture swap
   {
     int texturenum;

     in.read((char *) &texturenum, sizeof(texturenum) );
     thisface.numpoints = 0;
     thisface.points[0].x = (float) texturenum;
   }
   else
   {
      thisface.numpoints = numpoints;
      for(i = 0; i < numpoints; i++)
      {
         readCoordinate(thisface.points[i]);
         readCoordinate(thisface.normals[i]);
         in.read((char *) &thisface.texels[i].s, sizeof(float) );
         in.read((char *) &thisface.texels[i].t, sizeof(float) );
      }
   }
}


void BinaryMeshFile::initCollisionDetection(CollisionModel3D * collision_detection, Coordinate translation)
{
  int currentface;
  int j;

  for(currentface =0; currentface < numfaces+ numtextures - 1; currentface++)
  {
     Face & thisface = faces[currentface];

     if(thisface.numpoints == 3)
     {
        Coordinate pt0 = thisface.points[0] + translation;
        Coordinate pt1 = thisface.points[1] + translation;
        Coordinate pt2 = thisface.points[2] + translation;

        collision_detection->addTriangle(pt0.x, pt0.y, pt0.z, 
                                        pt1.x, pt1.y, pt1.z,
                                        pt2.x, pt2.y, pt2.z);
		for(j=0; j < 3; j++)
		{
		  maxminTriangle(thisface.points[j].x, thisface.points[j].y, thisface.points[j].z );
		}
     }
     else if(thisface.numpoints == 4)
     {
        Coordinate pt0 = thisface.points[0] + translation;
        Coordinate pt1 = thisface.points[1] + translation;
        Coordinate pt2 = thisface.points[2] + translation;
        Coordinate pt3 = thisface.points[3] + translation;

        collision_detection->addTriangle(pt0.x, pt0.y, pt0.z, 
                                        pt1.x, pt1.y, pt1.z,
                                        pt2.x, pt2.y, pt2.z);

        collision_detection->addTriangle(pt0.x, pt0.y, pt0.z, 
                                        pt2.x, pt2.y, pt2.z,
                                        pt3.x, pt3.y, pt3.z);


		for(j=0; j < 4; j++)
		{
		  maxminTriangle(thisface.points[j].x, thisface.points[j].y, thisface.points[j].z );
		}

     }
    
   }

}

CollisionModel3D * BinaryMeshFile::getCollisionModel()
{
  return mycollision_detection;
}

void BinaryMeshFile::setGreater(float testvalue, float & storto)
{
  if(testvalue > storto)
  {
    storto = testvalue;
  }
}

void BinaryMeshFile::setLess(float testvalue, float & storto)
{
  if(testvalue < storto)
  {
    storto = testvalue;
  }
}



void BinaryMeshFile::maxminTriangle(float x, float y, float z)
{  
  setLess(x, minx);
  setLess(y, miny);
  setLess(z, minz);
  setGreater(x, maxx);
  setGreater(y, maxy);
  setGreater(z, maxz);
}

void BinaryMeshFile::initExtents()
{
  Coordinate deltaextents;
  extents.x = maxx - minx;
  extents.y = maxy - miny;
  extents.z = maxz - minz;

  deltaextents = extents * .5f;

  
  center = Coordinate((maxx + minx) * .5f, 
					  (maxy + miny) * .5f, 
                      (maxz + minz) * .5f );


  radius = fastsqrt( fabs_f(deltaextents.x * deltaextents.x) + 
	             fabs_f(deltaextents.y * deltaextents.y) + 
                 fabs_f(deltaextents.z * deltaextents.z)  );
}


Coordinate BinaryMeshFile::getMins()
{
  return Coordinate(minx, miny, minz);
}


Coordinate BinaryMeshFile::getMaxs()
{
  return Coordinate(maxx, maxy, maxz);
}

void BinaryMeshFile::renderWireFrame(Color newcolor)
{
  int currentpoint;
  int currentface;
  bool lastwastriangle=true;
  int lasttexture = 0;
  int i;

  glDisable(GL_TEXTURE_2D);
  glEnable(GL_COLOR_MATERIAL);
  newcolor.render();
  currentpoint = 0;

  GLERRORCHECK

  for(currentface =0; currentface < numfaces+ numtextures - 1; currentface++)
  {
     Face & thisface = faces[currentface];

     if(thisface.numpoints == 3)
     {
        glBegin(GL_LINE_LOOP);
        for(i=0; i < 3; i++)
        {

	       glNormal3f(thisface.normals[i].x, thisface.normals[i].y, thisface.normals[i].z);
           glVertex3f(thisface.points[i].x , thisface.points[i].y , thisface.points[i].z);
        }
        glEnd();
     }
     else if(thisface.numpoints == 4)
     {
        glBegin(GL_LINE_LOOP);
        for(i=0; i < 4; i++)
        {
           glNormal3f(thisface.normals[i].x, thisface.normals[i].y, thisface.normals[i].z);
           glVertex3f(thisface.points[i].x , thisface.points[i].y , thisface.points[i].z);
        }
        glEnd();
     }
     else
     {

     }
  }
//  glEnd();
  GLERRORCHECK
}

void BinaryMeshFile::addToCollisionModel(CollisionModel3D * model, Coordinate translation)
{
  initCollisionDetection(model, translation);
}
