// Missile.cpp: implementation of the Missile class.
//
//////////////////////////////////////////////////////////////////////

#include "Missile.h"
#include "SeekBehavior.h"
#include "SoundFactory.h"
#include "Mission.h"
#include "SoundEffect3d.h"
#include "WorldOrientedBillboard.h"
#include "Camera.h"
#include "Mission.h"
#include "CockpitCamera.h"

#include "TextureFactory.h"
#include "PositionHistory.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Missile::Missile(mesh * newmesh,  Coordinate position, Coordinate newheading, Coordinate newup, Coordinate newleft,
        shipareas newarmor,
        float newvelocity, float newroll, float newpitch, float newyaw,
		float acceleration,
        Image * newimage, string newname, int newdamage, float newrange,
        bool tracking, bool torp, bool dturrets,
		SoundEffect3d * newsound,
		Texture * newtexture, Texture * exhausttexture,
		float newspoofprobab)
        : Targetable(newmesh, position, newheading, newup, newleft,
          shipareas(0,0, 0,0), newarmor, newvelocity, newroll, newpitch, newyaw,
		  acceleration, -1, //no angular acceleration available
		  newimage, newname), 
          damageamount(newdamage), damage_capships(torp), damage_turrets(dturrets),
		  sound(newsound),
		  spoofprobability(newspoofprobab),
		  is_spoofed(false)
{
  fired_by_player = false;
  setMaxDistance(newrange);
  if(tracking)
    setAI(& SeekBehavior::instance );
  setMaxVelocity();
  enginebillboard = new WorldOrientedBillboard(newtexture, 4.0f, 4.0f);
  enginebillboard->setCameraRelative(false);

  if(exhausttexture != NULL)
  {
    exhaustbillboard= new WorldOrientedBillboard(exhausttexture, 4.0f, 4.0f);
    exhaustbillboard->setCameraRelative(false);
  }
  else
  {
    exhaustbillboard = NULL;
  }
  rangesquared = newrange * newrange;
}

Missile::~Missile()
{
  delete enginebillboard;
  if(exhaustbillboard != NULL)
  {
    delete exhaustbillboard;
  }
}

Color Missile::getRadarColor(ship * player)
{
  return Color(.9f, .9f, 0.0f);
}

void Missile::onCollisionWith(Object3d * newobject)
{
  if(canDamage(newobject) )
  {
      newobject->damage(damageamount, getLocation() );
      if(fired_by_player)
      {
         Mission * mission = Mission::getCurrentMission();
         if( newobject->isdestroyed() && newobject->instanceof(INST_SHIP))
         {
            // we have a kill
            mission->incrementPlayerKills(); 
         }
         // player had a hit
         mission->incrementPlayerMissileHits();        
      }
  }
  SoundEffect3d * missilexplodesound = 
  SoundFactory::getSoundFactory()->getMissileExplodeSound();
  
  missilexplodesound->setPosition(getLocation() );
  missilexplodesound->setHeading(getHeading() );
  missilexplodesound->play();
  setDestroyed();
}

void Missile::damage(int amount, Coordinate fromlocation)
{
  setDestroyed();
}

void Missile::setCantHit(Object3d * other)
{
  donthit = other;
}

bool Missile::cantHit(Object3d * other)
{
  return other == donthit;
}

Targetable * Missile::getTarget()
{
  return target;
}

void Missile::render()
{
  Targetable::render();

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);

  if(exhaustbillboard != NULL)
  {
    renderExhaust();
  }

  enginebillboard->render();
  glDisable(GL_BLEND);
}

void Missile::targetDestroyed()
{
  target = NULL;
}

void Missile::setFiredByPlayer()
{
  fired_by_player = true;
}

void Missile::advance()
{
	bankleft();
	Targetable::advance();
	if(sound != NULL && sound->isPlaying() )
	{
    	sound->setPosition(getLocation() );
	    sound->render();
	}
	else
	{
	   delete sound;
       sound = NULL;	   
	}

	enginebillboard->setLocation(getcenter() + getHeading() * (getExtents().z * -.5f - .1f) );
	enginebillboard->setDesiredUp(getUp() );
	if(exhaustbillboard != NULL)
	{
	  exhaustbillboard->setDesiredUp( Camera::getCamera()->getUp()  );
	}
}


bool Missile::instanceof(char classname [ ])
{
  if(0 == strcmp(classname, "Missile")) 
    return true;
  return Targetable::instanceof(classname);
}	


void Missile::setDestroyed()
{
	Object3d::setDestroyed();

    if(this == Camera::getFocusObject() )
	{
		Camera::setFocusObject(Mission::getCurrentMission()->getPlayer() );
        Camera::setCamera(CockpitCamera::getInstance() );

	}
}

bool Missile::canDamage(Object3d * newobject)
{
  if(newobject->instanceof(INST_TURRET) && damage_turrets )
  {
    return true;
  }
  else if(newobject->instanceof(INST_COMPONENT) && damage_capships )
  {
    return true;
  }
  else if(newobject->instanceof(INST_TARGETABLE) && !(newobject->instanceof(INST_CAPITALSHIP)) )
  {
    return true;	  
  }
  else
  {
    return false;
  }
}


bool Missile::instanceof(InstanceT it)
{
  return it & (INST_OBJECT | INST_TARGETABLE | INST_MISSILE);
}

void Missile::renderExhaust()
{
  int i;
  int end =  history->getNumElements() - 1;
  float alpha;

  for(i=end; i >= 0; i--)
  {
	Coordinate location = history->getElement(i);
    exhaustbillboard->setLocation(location);
	if(end == 0.0f)
	{
	  alpha = 1.0f;
	}
	else
	{
	  alpha = ((float)(i)) / (float) end;	
	}
	exhaustbillboard->setAlpha(alpha);
	exhaustbillboard->render();
  }
}

void Missile::attemptSpoof(Coordinate decoylocation)
{
  if(is_spoofed)
	  return;
  if(spoofprobability == 0.0f)
	  return;

  float distancesquared = getLocation().distanceSquared(decoylocation);
  if(distancesquared > rangesquared)
  {
	return;
  }

  int randbase = rand();
  float randpercent = ( 
	  (float) rand()  / (float) RAND_MAX 
	  );
  if(randpercent < spoofprobability)
  {
    doSpoof(decoylocation);
  }

}

void Missile::doSpoof(Coordinate decoylocation)
{
  setTarget(NULL);
}
