// SettingsPanel.cpp: implementation of the SettingsPanel class.
//
//////////////////////////////////////////////////////////////////////
#include <GL/glew.h>
#ifdef WIN32
#include "gl/wglew.h"
#endif
#include "SettingsPanel.h"
#include "Settings.h"
#include "JoystickFactory.h"
#include "ForceFeedbackJoystick.h"
#include "TextureFactory.h"
#include "Math.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// res
#define RESOLUTIONLINE 0
// game details
#define NEBULALINE 1
#define COCKPITDETAILSLINE 2
#define DYNAMICLIGHTSLINE 3
#define REFLECTIONSLINE 4

// texture settings
#define SMOOTHINGLINE 5
#define MIPMAPLINE 6
#define COMPRESSLINE 7
#define ANISOTROPYLINE 8
#define VSYNCLINE 9

// audio
#define MUSICLINE 10
#define SOUNDLINE 11

// joystick
#define FORCELINE 12

#define NUMLINES 13


SettingsPanel::SettingsPanel()
{
  int i;

  Settings * settings = Settings::getSettings();


  resolution = settings->getResolutionString();
  smooth = settings->smooth();
  music = settings->getMusic();
  sound = settings->getSound();
  mipmap = settings->getMipMap();
  nebula = settings->getNebula();
  cockpitdetails = settings->getCockpitDetails();
  dynamiclights = settings->getDynamicLights();
  texturecompression = settings->getTextureCompression();
  anisotropy = settings->getAnisotropy();
  reflections = settings->getReflections();
  
#ifdef WIN32
  if(WGLEW_EXT_swap_control)
  {
    vsync = settings->getVSync();
  }
  else
  {
    vsync = true;
  }
#else
  vsync = true;
#endif

  modes = DisplayMode::getDisplayModes();
  
  // find the current mode
  DisplayMode currentmode(resolution);
  displayindex = 0;
  for(i=0; i < modes.size(); i++)
  {
    if(modes[i] == currentmode)
	{
       displayindex = i;
	}
  }
  
  
  setTitle("Settings");
  setNumLines(NUMLINES);
  reset();
  setConfirmText("OK");

  JoystickFactory * factory = JoystickFactory::getInstance();
  ForceFeedbackJoystick * ff = factory->getForceFeedbackJoystick();
  if(ff != NULL)
  {
      force = "Force Feedback Enabled";
  }
  else
  {
	  Joystick * joy = factory->getJoystick();
	  if(joy != NULL)
	  {
        force = "Force Feedback Disabled";
	  }
	  else
	  {
        force = "Force Feedback Disabled";
	  }
  }
  JoystickFactory::getInstance()->clear();
  setLine(force, FORCELINE);
}

SettingsPanel::~SettingsPanel()
{

}


void SettingsPanel::onLineClick(int linenumber)
{
  if(linenumber == NEBULALINE)
  {
    nebula = ! nebula;
  }
  else if(linenumber == MIPMAPLINE)
  {
    mipmap = ! mipmap;
  }
  else if(linenumber == SOUNDLINE)
  {
    sound = ! sound;
  }
  else if(linenumber == MUSICLINE)
  {
    music = ! music;
  }
  else if(linenumber == SMOOTHINGLINE)
  {
    smooth = ! smooth;
  }
  else if(linenumber == COCKPITDETAILSLINE)
  {
    cockpitdetails = ! cockpitdetails;
  }

  else if(linenumber == RESOLUTIONLINE)
  {
	displayindex = (displayindex + 1) % modes.size();
	resolution = modes[displayindex].toString();
  }
  else if(linenumber == DYNAMICLIGHTSLINE)
  {
    dynamiclights = ! dynamiclights;
  }
  else if(linenumber == COMPRESSLINE)
  {
     int max = TextureFactory::getMaxTextureCompression();

	 texturecompression = (texturecompression + 1) % (max + 1) ;
	 texturecompression = MIN(texturecompression,  max);
  }
  else if(linenumber == ANISOTROPYLINE)
  {
	anisotropy = anisotropy << 1;
	if(anisotropy > TextureFactory::getMaxAnisotropy() )
	{
      anisotropy = 1;
	}

  }
  else if(linenumber == VSYNCLINE)
  {
#ifdef WIN32
	if(WGLEW_EXT_swap_control)
	{
		vsync = ! vsync;
	}
#endif
  }
  else if(linenumber == REFLECTIONSLINE)
  {
    if(GLEW_EXT_texture_cube_map)
	{
		reflections = !reflections;
	}
  }
  reset();
}

void SettingsPanel::onCancel()
{
  int i;

  Settings * settings = Settings::getSettings();

  resolution = settings->getResolutionString();
  DisplayMode currentmode(resolution);
  displayindex = 0;
  for(i=0; i < modes.size(); i++)
  {
    if(modes[i] == currentmode)
	{
       displayindex = i;
	}
  }

  smooth = settings->smooth();
  music = settings->getMusic();
  sound = settings->getSound();
  mipmap = settings->getMipMap();
  nebula = settings->getNebula();
  texturecompression = settings->getTextureCompression();
  anisotropy = settings->getAnisotropy();
  cockpitdetails = settings->getCockpitDetails();
  vsync = settings->getVSync();
  if(GLEW_EXT_texture_cube_map)
  {
    reflections = settings->getReflections();
  }
  else
  {
	reflections = false;
  }
  reset();
}

void SettingsPanel::onConfirm()
{
  Settings * settings = Settings::getSettings();

  settings->setSmooth(smooth);
  settings->setResolutionString(resolution  );
  settings->setMusic(music);
  settings->setSound(sound);
  settings->setMipMap(mipmap);
  settings->setNebula(nebula);
  settings->setCockpitDetails(cockpitdetails);
  settings->setDynamicLights(dynamiclights);
  settings->setTextureCompression(texturecompression);
  settings->setAnisotropy(anisotropy);
  settings->setVSync(vsync);
  settings->setReflections(reflections);
  settings->write();
}

void SettingsPanel::reset()
{
  setLine(force, FORCELINE);
  setLine(resolution , RESOLUTIONLINE);
  if(smooth)
    setLine("Smooth lines (nicer but slower)", SMOOTHINGLINE);
  else
    setLine("Jagged lines (faster)", SMOOTHINGLINE);
  if(music)
    setLine("Music on", MUSICLINE);
  else
    setLine("Music off", MUSICLINE);
  if(sound)
    setLine("Sound on", SOUNDLINE);
  else
    setLine("Sound off", SOUNDLINE);
  if(mipmap)
    setLine("Mip-mapping on (nicer but slower)", MIPMAPLINE);
  else
    setLine("Mip-mapping off (faster)", MIPMAPLINE);
  if(nebula)
    setLine("Nebula on", NEBULALINE);
  else
    setLine("Nebula off", NEBULALINE);
  if(cockpitdetails)
    setLine("Cockpit details on", COCKPITDETAILSLINE);
  else
    setLine("Cockpit details off", COCKPITDETAILSLINE);
  if(dynamiclights)
    setLine("Dynamic lights on", DYNAMICLIGHTSLINE);
  else
    setLine("Dynamic lights off", DYNAMICLIGHTSLINE);

  if(texturecompression == 0)
  {
    setLine("Texture compression off (nicer but slower)", COMPRESSLINE);
  }
  else if(texturecompression == 1)
  {
    setLine("Basic texture compression on (faster)", COMPRESSLINE);
  }
  else if(texturecompression == 2)
  {
    setLine("Advanced texture compression on (faster)", COMPRESSLINE);
  }

  char buffer[100];
  if(anisotropy == 1)
  {
     sprintf(buffer, "Texture sharpness 1 (fastest, least nice)" );
  }
  else if(anisotropy == TextureFactory::getMaxAnisotropy() )
  {
     sprintf(buffer, "Texture sharpness %d (slowest, nicest)", anisotropy);
  }
  else
  {
     sprintf(buffer, "Texture sharpness %d", anisotropy);
  }
  setLine(buffer, ANISOTROPYLINE);
  if(vsync)
  {
    setLine("Vertical sync enabled (slower)", VSYNCLINE);	
  }
  else
  {
    setLine("Vertical sync disabled (faster)", VSYNCLINE);	
  }

  if(reflections)
  {
    setLine("Reflections on", REFLECTIONSLINE);	
  }
  else
  {
    setLine("Reflections off", REFLECTIONSLINE);	
  }

}
